

#include <pic.h>
#include "main677.h"

__CONFIG(INTIO & WDTDIS & PWRTEN & MCLRDIS & UNPROTECT & BORDIS & IESODIS & FCMDIS);


/** Variables **/

// Debounce Variables

unsigned char	DebounceCount;
unsigned char	Last;
unsigned char	Pressed;

FFlags Flags;							// Flag variable with SSP busy flag bit.
BButtons Buttons;						// Flags for capacitive sensing, indicates a button down

// I2C Communications
unsigned char	ByteCount;	
unsigned int	TempInt;
unsigned char	SSPCommand;
unsigned char	SSPIndex;
unsigned char	SSPSize;
unsigned char	IndexCount;


// Capacitive Variables
long			percent;
unsigned char	AvgIndex;

unsigned int	RAW	   [NUM_BTTNS];										// Variables for CapTouch and communication standards to GUI
unsigned int	AVERAGE[NUM_BTTNS];
unsigned int	TRIP   [NUM_BTTNS];										// Trip & guard in this 16f677 application
unsigned int	GUARD  [NUM_BTTNS];										// are just used to communicate to the mtouch gui

unsigned char	pctarray[NUM_BTTNS];									// Percent array (like trip for absolute projects), holds individual pct thresholds

unsigned char	FIRST;													// first variable to 'discard' first N samples
unsigned char	INDEX;													// Index of what button is being checked.

unsigned int	VALUE;													// current button value
unsigned int	BIGVAL;
unsigned int	SMALLAVG;


/** Constants **/

// 0x94/0xA0 = Cin0-    ,   0x95/0xA1 = Cin1-   (Sets comparator operation and proper channel.)
const unsigned char	COMP1[4] = {0x94, 0x95, 0x96, 0x97};				// comparator preset values for each button
const unsigned char	COMP2[4] = {0xA0, 0xA1, 0xA2, 0xA3};





/*....................................................................
. 	main()
.
.		Entry point of program.
....................................................................*/
void main(void) {

	Init();									// Perform initializations
	
	while(1) {								// Main Loop Code


		// Set LEDs accordingly				(restricted to 1 at a time in isr)
		switch (Pressed) {
			case NVAL:		LED_N  = ON; break;
			case NEVAL:		LED_NE = ON; break;
			case EVAL:		LED_E  = ON; break;	// (if : else)
			case SEVAL:		LED_SE = ON; break;
			case SVAL:		LED_S  = ON; break;
			case SWVAL:		LED_SW = ON; break;
			case WVAL:		LED_W  = ON; break;
			case NWVAL:		LED_NW = ON; break;
			case ALLOFF:	LED_N = LED_NE = LED_E = LED_SE = OFF;
							LED_S = LED_SW = LED_W = LED_NW = OFF; break;
			default:		break;
		}


		// For mTouch Diagnostic Tool, set up for __read only__
		TRIP[INDEX]  = (unsigned long)AVERAGE[INDEX]*pctarray[INDEX]/1000; 		// Compute trip values in absolute form, calc from pct full-on (single button press), PCT_HALFON = weaker press
		GUARD[INDEX] = (unsigned long)AVERAGE[INDEX]*PCT_AVG/1000;			// Set Guard eq. to full off value, in absolute form.

	}

}





/*....................................................................
.	Init()
.
.		Initializations performed on power-up.
....................................................................*/
void Init(void) {

	// Pin I/O Direction
	// Cap and SSP I/O set in respective functions

	// LEDs
	TRISB5 = 0;							// N	Set LED pins as outputs.
	TRISA2 = 0;							// NE
	TRISA0 = 0;							// E
	TRISA4 = 0;							// SE
	TRISC5 = 0;							// S
	TRISC6 = 0;							// SW
	TRISC7 = 0;							// W
	TRISB7 = 0;							// NW

	// Application Initializations
	CapInit();
	SSPInit();

	LED_N  = OFF;						// Clear all LEDs at startup
	LED_NE = OFF;						// (after ANSEL's, and appropriate cap/ssp init's)
	LED_E  = OFF;
	LED_SE = OFF;
	LED_S  = OFF;
	LED_SW = OFF;
	LED_W  = OFF;
	LED_NW = OFF;



	// Capacitive Paired-press Debounce setup
	Last    = ALLOFF;					// No Last button or pressed button at startup
	Pressed = ALLOFF;					// (init them off)
	DebounceCount  = 0;
	Flags.BTNVALID = 0;					// No button is valid on powerup
	Flags.AVERAGESTOP = 0;				// Don't stop averaging on powerup

	GIE = 1;							// Enable All Interrupts
}




/*....................................................................
.	CapInit()
.
.		Initializes device for capacitive needs.
....................................................................*/
void CapInit(void) {

	// Capacitive Pins
	// Cap needs to set up C2OUT as o/p, and T1CKI=i/p, and C12INx- pins used = i/p.
	TRISC0 = 1;									// C2IN+ (input Volt ref.)
	TRISA1 = 1;									// C12IN0-
	TRISC1 = 1;									// C12IN1- Comparator inputs connect to pads
	TRISC2 = 1;									// C12IN2-
	TRISC3 = 1;									// C12IN3-
	TRISA5 = 1;									// T1CKI as input
	TRISC4 = 0;									// C2OUT = output.


	// Load defaults for buttons
	for (INDEX=0; INDEX < NUM_BTTNS; INDEX++) {
		RAW[INDEX] = 0;							// Init raw and avg to zero
		AVERAGE[INDEX] = 0;
		TRIP[INDEX] = 0;						// TRIP  - USED ONLY FOR FEEDBACK IN MAIN LOOP CODE
		GUARD[INDEX] = 0;						// GUARD - USED ONLY FOR FEEDBACK IN MAIN LOOP CODE
		pctarray[INDEX] = PCT_FULLON;			// Init sensors to default sensor threshold
		
		// Individually Adjusted Thresholds
		// ...
		// (none)
	
		// Note: HALFON thresholds are used exclusively in the service routine.

	}


	FIRST = 60;									// First pass flag to stabilize average on powerup

	ANSEL   = 0xF2;								// Configure button inputs analog, along with the 1/4VDD reference input, all else digital
	ANSELH  = 0x00;								// ANS<11:8> as digital
	
	OSCCON	= 0x71;								// (Optional) Use 8MHz internal clock

	INDEX	= 0;								// Set Index for first channel (Button state machine state variable.)
	CM1CON0 = COMP1[INDEX];						// Preset the comparators for first channel 
	CM2CON0 = COMP2[INDEX];						// .. (and set appropriate enable bits)
	CM2CON1 = 0x00;

	VRCON   = 0xCF;								// setup the CVref for 2/3VDD
	SRCON   = 0xF0;								// turn on the SR flip flop mode

	T1CON	= 0x06;								// enable timer1
	OPTION	= 0x83;								// setup the timer0 prescaler (3 LSb's)

	RestartTimers();							// Clear/Start for first time.

	INTCON	= 0;								// clear the interrupt register
	T0IE	= 1;								// turn on the timer0 interrupt

}




/*....................................................................
.	RestartTimers()
.
.		Resets and restarts timers 0 & 1 used for capacitive sensing.
....................................................................*/
void RestartTimers(void) {

	TMR1L	= 0;								// reset Timer1
	TMR1H	= 0;								//   "     "
	TMR1ON	= 1;								// restart Timer1
	TMR0	= 0;								// reset Timer0
	T0IF	= 0;								// clear the interrupt
}




/*....................................................................
.	SetNextChannel()
.
.		Performs necessary tasks to set next sensor to scan.
....................................................................*/
void SetNextChannel(void) {

	INDEX   = (++INDEX) & 0x03;					// Incr. index with rollover
	CM1CON0 = COMP1[INDEX];						// Set comparator inputs
	CM2CON0 = COMP2[INDEX];						// Set comparator inputs
}




/*....................................................................
.	isr()
.
.		Interrupt Service Routine - is the function called when any
.	interrupt is received on the part.  The device context saves, 
.	and then redirects flow here. Program flow will be returned at
.	the end of this function.  The flagging interrupt MUST be
.	cleared in software! (or you will be stuck for-ev-er).
....................................................................*/
static void interrupt isr(void)
{																		
	
	/* Timer 0 Interrupt */
	/*********************/
	if (T0IF == 1) {														// Timer 0 Overflow Occurred, 
		// Service Capacitive Routine
		// .. pending does not interfere with I2C
		TMR1ON	= 0;														// stop Timer1
	
		if (Flags.SSPBF == 1)												// SSP is busy, and can not interrupt it.
			RestartTimers();												// Clear timers
		else 																// SSP is not busy, do cap. stuff
			CapISR();														// Service Routine {resets/restarts tmrs}			
	}


	/* SSP Interrupt */
	/*****************/
	if (SSPIF == 1)															// SSP Interrupt Detected
		SSPISR();
	

	/* Timer Interrupt During SSP Intpt? */
	/*************************************/
	if (T0IF == 1)															// Catch if timer interrupt during ssp?  Handled at end of ISR.
		RestartTimers();													// (Clears T0IF)

}





/*....................................................................
.	CapISR()
.
.		Capacitive Interrupt Service Routine performs detection of
.	button presses and is called in response to a T0IF.  It is also
.	responsible for clearing the T0IF flag.
....................................................................*/
void CapISR(void) {
	

	// i.
	if (Flags.SLEEP) {
		return;																// During sleep, some code is run when checking to wake up from ssp.  Prevent anything from happening.
	}
	
	// 1. Get the raw sensor reading.

	// ...
	VALUE = TMR1L + (unsigned int)(TMR1H << 8);								// Get the frequency count for the current sensor
	BIGVAL   = VALUE * 16;													// BIGVAL is current measurement left shifted 4 bits
	SMALLAVG = AVERAGE[INDEX] / 16;											// SMALLAVG is the current average right shifted 4 bits
	RAW[INDEX] = BIGVAL;													// RAW array holds the most recent BIGVAL values

	// 2. On power-up, reach steady-state readings

	// ...	
	// During power-up, system must be initialized, set average ea. pass
	if (FIRST > 0) {
		FIRST--;															// Decr. N # times to establish startup
		AVERAGE[INDEX] = BIGVAL;											// During start up time, set Average each pass.
		SetNextChannel();
		RestartTimers();													// Restart the timers
		return;																// .. and exit
	}


	// 3. Compute Percent Shift
	
	// ...
	percent = ((long)AVERAGE[INDEX] - (long)BIGVAL);						// Need to cast AVG & RAW or result is not negative.
	if (percent < 0){														// If negative set to 0 (Means RAW > AVG, either release or steady state avg error)
		percent = 0;														// It is easiest to deal with negatives here rather than later if statements. 
	} else {
		percent = percent * 1000;											//              (AVG - RAW)  *1000
		percent = percent / AVERAGE[INDEX];									// now div by:        AVG	     	= percent (whole percents w/ first decimal, no fractions)
	}



	// 4. Is a button pressed?

	// ...	
	//	Set button flags if sensor pressed
	//  Full Button Flags
	if (percent > pctarray[INDEX]) {
		switch(INDEX) {
			case 0:	Buttons.BTN0 = 1; break;							
			case 1:	Buttons.BTN1 = 1; break;							
			case 2:	Buttons.BTN2 = 1; break;
			case 3:	Buttons.BTN3 = 1; break;
			default: break;
		}
	} else if (percent < PCT_FULLOFF) {
		switch(INDEX) {
			case 0:	Buttons.BTN0 = 0; break;
			case 1:	Buttons.BTN1 = 0; break;
			case 2:	Buttons.BTN2 = 0; break;
			case 3:	Buttons.BTN3 = 0; break;
			default: break;
		}
	}


	// Half Button Flags (Paired presses detect at weaker threshold)
	if (percent > PCT_HALFON) {
		switch(INDEX) {
			case 0:	Buttons.BTN0H = 1; break;
			case 1:	Buttons.BTN1H = 1; break;
			case 2:	Buttons.BTN2H = 1; break;
			case 3:	Buttons.BTN3H = 1; break;
			default: break;
		}
	} else if (percent < PCT_HALFOFF) {
		switch(INDEX) {
			case 0:	Buttons.BTN0H = 0; break;
			case 1:	Buttons.BTN1H = 0; break;
			case 2:	Buttons.BTN2H = 0; break;
			case 3:	Buttons.BTN3H = 0; break;
			default: break;
		}
	}



	// 	Using Paired Presses, set buttons after scanning all 4 sensors as one pass

	if (INDEX == 3) {
		// At end of scanning all, check for presses
		
		Buttons.N  = 0;							// Clear all to start
		Buttons.NE = 0;	
		Buttons.E  = 0;							// .. next will set appropriately
		Buttons.SE = 0;
		Buttons.S  = 0;
		Buttons.SW = 0;
		Buttons.W  = 0;
		Buttons.NW = 0;

		// First do the paired presses, NE, SE, SW, NW
		//       0			01 = NE
		//   2       1		13 = SE
		//       3			23 = SW
		//					02 = NW
		if (Buttons.BTN0H && Buttons.BTN1H) {				
			Buttons.NE = 1;
			Last = NEVAL;
		} else if (Buttons.BTN1H && Buttons.BTN3H) {
			Buttons.SE = 1;
			Last = SEVAL;
		} else if (Buttons.BTN2H && Buttons.BTN3H) {
			Buttons.SW = 1;
			Last = SWVAL;
		} else if (Buttons.BTN0H && Buttons.BTN2H) {
			Buttons.NW = 1;
			Last = NWVAL;
		} else if (Buttons.BTN0) {
			Buttons.N = 1;
			Last = NVAL;
		} else if (Buttons.BTN1) {
			Buttons.E = 1;
			Last = EVAL;
		} else if (Buttons.BTN3) {
			Buttons.S = 1;
			Last = SVAL;
		} else if (Buttons.BTN2) {
			Buttons.W = 1;
			Last = WVAL;
		} else {
			Last = ALLOFF;
		}

	

		// Provide some debounce
		// ...
		if (Last != Pressed && Pressed != ALLOFF) {						// A BUTTON THAT'S NOT THE PRESSED BUTTON IS BEING TOUCHED
			// Decrement debouncer	
			if (--DebounceCount < 1)		DebounceCount = 1;			// Lower limit of counter
			if (DebounceCount == 1) {
				Flags.BTNVALID = 0;										// At lwr lmt, mark button flags as invalid.
				Pressed = ALLOFF;										// must indicate when no buttons are pressed
			}

		} else if (Last == Pressed) {									// THE BUTTON PRESSED IS STILL BEING TOUCHED
			// Increment debouncer
			DebounceCount += 2;
			if (Last == NEVAL || Last == SEVAL || Last == SWVAL || Last == NWVAL)
				DebounceCount++;
			else
				DebounceCount--;

			if (DebounceCount > DBMAX) 		DebounceCount = DBMAX;		// Upper limit of counter
			if (DebounceCount == DBMAX)		{
				Flags.BTNVALID = 1;										// Once reached limit, set global flag for btn as debounced and valid
				Pressed = Last;											// If activating a button as pressed, store it in this variable for comparison for debounce use
			}

		} else if (Last != ALLOFF && Pressed == ALLOFF) {				// A BUTTON IS PRESSED, AND NO BUTTON INDICATES IN Pressed VARIABLE
			// Increment debouncer
			DebounceCount += 2;
			if (Last == NEVAL || Last == SEVAL || Last == SWVAL || Last == NWVAL)
				DebounceCount++;
			else
				DebounceCount--;

			if (DebounceCount > DBMAX) 		DebounceCount = DBMAX;		// Upper limit of counter
			if (DebounceCount == DBMAX)		{
				Flags.BTNVALID = 1;										// Once reached limit, set global flag for btn as debounced and valid
				Pressed = Last;											// If activating a button as pressed, store it in this variable for comparison for debounce use
			}
			
		} else if (Last == ALLOFF) {									// ALL BUTTONS ARE OFF
			// Decrement debouncer
			if (--DebounceCount < 1)		DebounceCount = 1;			// Lower limit of counter
			if (DebounceCount == 1) {
				Flags.BTNVALID = 0;										// At lwr lmt, mark button flags as invalid.
				Pressed = ALLOFF;										// must indicate when no buttons are pressed
			}

		}
		
	}



	// 5. Implement quick-release for a released button

	// ...
	if (BIGVAL > AVERAGE[INDEX]) {
		// Average upwards to higher raw value (but no timed delay)
		// (Averaging protects against flukes, but takes longer to restablish average)
		AVERAGE[INDEX] = AVERAGE[INDEX] + (VALUE - SMALLAVG);
	}


	// 6. Average in the new value

	// ...
	// Determine if i should average in the new value
	Flags.AVERAGESTOP = Flags.BTNVALID;									// If a button is valid, stop averaging activity, else allow averaging (it's a renamed separate bit flag equals btnvalid flag)


	// Do average on periodic schedule
	if ((percent < PCT_AVG) && !Flags.AVERAGESTOP) { 					// Use lesser value of 'off', half-off, for averaging limit
		if (AvgIndex < AVG_DELAY)			AvgIndex++;
		else								AvgIndex = 0;
		
		if (AvgIndex == AVG_DELAY)
			AVERAGE[INDEX] = AVERAGE[INDEX] + (VALUE - SMALLAVG);
			// Do the Average on time at Slow Rate
	}


	// Determine next sensor to test.
	SetNextChannel();

	// Restart timers for next pass
	RestartTimers();	
}



